import dagster_pandas as dagster_pd
from dagster_dask import DataFrame, dask_executor

from dagster import (
    InputDefinition,
    ModeDefinition,
    execute_pipeline,
    file_relative_path,
    pipeline,
    reconstructable,
    seven,
    solid,
)
from dagster.core.definitions.executor import default_executors
from dagster.core.definitions.reconstructable import ReconstructablePipeline
from dagster.core.instance import DagsterInstance
from dagster.core.test_utils import nesting_composite_pipeline


@solid
def simple(_):
    return 1


@pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])])
def dask_engine_pipeline():
    return simple()


def test_execute_on_dask_local():
    with seven.TemporaryDirectory() as tempdir:
        result = execute_pipeline(
            reconstructable(dask_engine_pipeline),
            run_config={
                'storage': {'filesystem': {'config': {'base_dir': tempdir}}},
                'execution': {'dask': {'config': {'cluster': {'local': {'timeout': 30}}}}},
            },
            instance=DagsterInstance.local_temp(),
        )
        assert result.result_for_solid('simple').output_value() == 1


def dask_composite_pipeline():
    return nesting_composite_pipeline(
        6, 2, mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])]
    )


def test_composite_execute():
    result = execute_pipeline(
        reconstructable(dask_composite_pipeline),
        run_config={
            'storage': {'filesystem': {}},
            'execution': {'dask': {'config': {'cluster': {'local': {'timeout': 30}}}}},
        },
        instance=DagsterInstance.local_temp(),
    )
    assert result.success


@solid(input_defs=[InputDefinition('df', dagster_pd.DataFrame)])
def pandas_solid(_, df):  # pylint: disable=unused-argument
    pass


@pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])])
def pandas_pipeline():
    return pandas_solid()


def test_pandas_dask():
    run_config = {
        'solids': {
            'pandas_solid': {
                'inputs': {'df': {'csv': {'path': file_relative_path(__file__, 'ex.csv')}}}
            }
        }
    }

    result = execute_pipeline(
        ReconstructablePipeline.for_file(__file__, pandas_pipeline.name),
        run_config={
            'storage': {'filesystem': {}},
            'execution': {'dask': {'config': {'cluster': {'local': {'timeout': 30}}}}},
            **run_config,
        },
        instance=DagsterInstance.local_temp(),
    )

    assert result.success


@solid(input_defs=[InputDefinition('df', DataFrame)])
def dask_solid(_, df):  # pylint: disable=unused-argument
    pass


@pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])])
def dask_pipeline():
    return dask_solid()


def test_dask():
    run_config = {
        'solids': {
            'dask_solid': {
                'inputs': {'df': {'csv': {'path': file_relative_path(__file__, 'ex*.csv')}}}
            }
        }
    }

    result = execute_pipeline(
        ReconstructablePipeline.for_file(__file__, dask_pipeline.name),
        run_config={
            'storage': {'filesystem': {}},
            'execution': {'dask': {'config': {'cluster': {'local': {'timeout': 30}}}}},
            **run_config,
        },
        instance=DagsterInstance.local_temp(),
    )

    assert result.success



import dask.dataframe as dd

from dagster import (
    Any,
    AssetMaterialization,
    Bool,
    DagsterInvariantViolationError,
    DagsterType,
    Enum,
    EnumValue,
    EventMetadataEntry,
    Field,
    Int,
    Permissive,
    String,
    TypeCheck,
    check,
    dagster_type_loader,
    dagster_type_materializer,
)
from dagster.config.field_utils import Selector

WriteCompressionTextOptions = Enum(
    'WriteCompressionText', [EnumValue('gzip'), EnumValue('bz2'), EnumValue('xz'),],
)

EngineParquetOptions = Enum(
    'EngineParquet', [EnumValue('auto'), EnumValue('fastparquet'), EnumValue('pyarrow'),],
)


def dict_without_keys(ddict, *keys):
    return {key: value for key, value in ddict.items() if key not in set(keys)}


@dagster_type_materializer(
    Selector(
        {
            'csv': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="str or list, Path glob indicating the naming scheme for the output files",
                    ),
                    'single_file': Field(
                        Bool,
                        is_required=False,
                        description="""
                            Whether to save everything into a single CSV file.
                            Under the single file mode, each partition is appended at the end of the specified CSV file.
                            Note that not all filesystems support the append mode and thus the single file mode,
                            especially on cloud storage systems such as S3 or GCS.
                            A warning will be issued when writing to a file that is not backed by a local filesystem.
                        """,
                    ),
                    'encoding': Field(
                        String,
                        is_required=False,
                        description="""
                            A string representing the encoding to use in the output file,
                            defaults to 'ascii' on Python 2 and 'utf-8' on Python 3.
                        """,
                    ),
                    'mode': Field(
                        String, is_required=False, description="Python write mode, default 'w'",
                    ),
                    'compression': Field(
                        WriteCompressionTextOptions,
                        is_required=False,
                        description="""
                            a string representing the compression to use in the output file,
                            allowed values are 'gzip', 'bz2', 'xz'.
                        """,
                    ),
                    'compute': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If true, immediately executes.
                            If False, returns a set of delayed objects, which can be computed at a later time.
                        """,
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="Parameters passed on to the backend filesystem class.",
                    ),
                    'header_first_partition_only': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If set to `True`, only write the header row in the first output file.
                            By default, headers are written to all partitions
                            under the multiple file mode (`single_file` is `False`)
                            and written only once under the single file mode (`single_file` is `True`).
                            It must not be `False` under the single file mode.
                        """,
                    ),
                    'compute_kwargs': Field(
                        Permissive(),
                        is_required=False,
                        description="Options to be passed in to the compute method",
                    ),
                }
            ),
            'parquet': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or pathlib.Path, Destination directory for data.
                            Prepend with protocol like ``s3://`` or ``hdfs://`` for remote data.
                        """,
                    ),
                    'engine': Field(
                        EngineParquetOptions,
                        is_required=False,
                        description="""
                            {'auto', 'fastparquet', 'pyarrow'}, default 'auto' Parquet library to use.
                            If only one library is installed, it will use that one; if both, it will use 'fastparquet'.
                        """,
                    ),
                    'compression': Field(
                        Any,
                        is_required=False,
                        description="""
                        str or dict, optional Either a string like ``'snappy'``
                        or a dictionary mapping column names to compressors like ``{'name': 'gzip', 'values': 'snappy'}``.
                        The default is ``'default'``, which uses the default compression for whichever engine is selected.
                        """,
                    ),
                    'write_index': Field(
                        Bool,
                        is_required=False,
                        description="Whether or not to write the index. Defaults to True.",
                    ),
                    'append': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If False (default), construct data-set from scratch.
                            If True, add new row-group(s) to an existing data-set.
                            In the latter case, the data-set must exist, and the schema must match the input data.
                        """,
                    ),
                    'ignore_divisions': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If False (default) raises error when previous divisions overlap with the new appended divisions.
                            Ignored if append=False.
                        """,
                    ),
                    'partition_on': Field(
                        list,
                        is_required=False,
                        description="""
                            Construct directory-based partitioning by splitting on these fields values.
                            Each dask partition will result in one or more datafiles, there will be no global groupby.
                        """,
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="Key/value pairs to be passed on to the file-system backend, if any.",
                    ),
                    'write_metadata_file': Field(
                        Bool,
                        is_required=False,
                        description="Whether to write the special '_metadata' file.",
                    ),
                    'compute': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If True (default) then the result is computed immediately.
                            If False then a ``dask.delayed`` object is returned for future computation.
                        """,
                    ),
                    'compute_kwargs': Field(
                        Permissive(),
                        is_required=False,
                        description="Options to be passed in to the compute method.",
                    ),
                }
            ),
            'hdf': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or pathlib.Path, Path to a target filename.
                            Supports strings, ``pathlib.Path``, or any object implementing the ``__fspath__`` protocol.
                            May contain a ``*`` to denote many filenames.
                        """,
                    ),
                    'key': Field(
                        String,
                        is_required=True,
                        description="""
                            Datapath within the files.
                            May contain a ``*`` to denote many locations.
                        """,
                    ),
                    'compute': Field(
                        Bool,
                        is_required=False,
                        description="""
                            Whether or not to execute immediately.
                            If False then this returns a ``dask.Delayed`` value.
                        """,
                    ),
                    'scheduler': Field(
                        String,
                        is_required=False,
                        description="The scheduler to use, like 'threads' or 'processes'.",
                    ),
                }
            ),
            'json': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or list, Location to write to.
                            If a string, and there are more than one partitions in df,
                            should include a glob character to expand into a set of file names,
                            or provide a ``name_function=`` parameter.
                            Supports protocol specifications such as ``'s3://'``.
                        """,
                    ),
                    'encoding': Field(
                        String,
                        is_required=False,
                        description="default is 'utf-8', The text encoding to implement, e.g., 'utf-8'.",
                    ),
                    'errors': Field(
                        String,
                        is_required=False,
                        description="default is 'strict', how to respond to errors in the conversion (see ``str.encode()``).",
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="Passed to backend file-system implementation",
                    ),
                    'compute': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If true, immediately executes.
                            If False, returns a set of delayed objects, which can be computed at a later time.
                        """,
                    ),
                    'compute_kwargs': Field(
                        Permissive(),
                        is_required=False,
                        description="Options to be passed in to the compute method",
                    ),
                    'compression': Field(
                        String, is_required=False, description="String like 'gzip' or 'xz'.",
                    ),
                },
            ),
            'sql': Permissive(
                {
                    'name': Field(String, is_required=True, description="Name of SQL table",),
                    'uri': Field(
                        String,
                        is_required=True,
                        description="Full sqlalchemy URI for the database connection",
                    ),
                    'schema': Field(
                        String,
                        is_required=False,
                        description="Specify the schema (if database flavor supports this). If None, use default schema.",
                    ),
                    'if_exists': Field(
                        String,
                        is_required=False,
                        description="""
                            {'fail', 'replace', 'append'}, default 'fail'"
                            How to behave if the table already exists.
                            * fail: Raise a ValueError.
                            * replace: Drop the table before inserting new values.
                            * append: Insert new values to the existing table.
                        """,
                    ),
                    'index': Field(
                        Bool,
                        is_required=False,
                        description="""
                            default is True, Write DataFrame index as a column.
                            Uses `index_label` as the column name in the table.
                        """,
                    ),
                    'index_label': Field(
                        Any,
                        is_required=False,
                        description="""
                            str or sequence, default None Column label for index column(s).
                            If None is given (default) and `index` is True, then the index names are used.
                            A sequence should be given if the DataFrame uses MultiIndex.
                        """,
                    ),
                    'chunksize': Field(
                        Int,
                        is_required=False,
                        description="""
                            Specify the number of rows in each batch to be written at a time.
                            By default, all rows will be written at once.
                        """,
                    ),
                    'dtype': Field(
                        Any,
                        is_required=False,
                        description="""
                            dict or scalar, Specifying the datatype for columns.
                            If a dictionary is used, the keys should be the column names
                            and the values should be the SQLAlchemy types or strings for the sqlite3 legacy mode.
                            If a scalar is provided, it will be applied to all columns.
                        """,
                    ),
                    'method': Field(
                        String,
                        is_required=False,
                        description="""
                            {None, 'multi', callable}, default None
                            Controls the SQL insertion clause used:
                            * None : Uses standard SQL ``INSERT`` clause (one per row).
                            * 'multi': Pass multiple values in a single ``INSERT`` clause.
                            * callable with signature ``(pd_table, conn, keys, data_iter)``.
                            Details and a sample callable implementation can be found in the
                            section :ref:`insert method <io.sql.method>`.
                        """,
                    ),
                    'compute': Field(
                        Bool,
                        is_required=False,
                        description="""
                            default is True, When true, call dask.compute and perform the load into SQL;
                            otherwise, return a Dask object (or array of per-block objects when parallel=True).
                        """,
                    ),
                    'parallel': Field(
                        Bool,
                        is_required=False,
                        description="""
                            default is False, When true, have each block append itself to the DB table concurrently.
                            This can result in DB rows being in a different order than the source DataFrame's corresponding rows.
                            When false, load each block into the SQL DB in sequence.
                        """,
                    ),
                },
            ),
        },
    )
)
def dataframe_materializer(_context, config, dask_df):
    check.inst_param(dask_df, 'dask_df', dd.DataFrame)
    file_type, file_options = list(config.items())[0]
    path = file_options.get('path')

    if file_type == 'csv':
        dask_df.to_csv(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'parquet':
        dask_df.to_parquet(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'hdf':
        dask_df.to_hdf(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'json':
        dask_df.to_json(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'sql':
        dask_df.to_sql(**file_options)
    else:
        check.failed('Unsupported file_type {file_type}'.format(file_type=file_type))

    return AssetMaterialization.file(path)


@dagster_type_loader(
    Selector(
        {
            'csv': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or list, Absolute or relative filepath(s).
                            Prefix with a protocol like `s3://` to read from alternative filesystems.
                            To read from multiple files you can pass a globstring or a list of paths,
                            with the caveat that they must all have the same protocol.
                        """,
                    ),
                    'blocksize': Field(
                        Any,
                        is_required=False,
                        description="""
                        str or int or None, Number of bytes by which to cut up larger files.
                        Default value is computed based on available physical memory and the number of cores, up to a maximum of 64MB.
                        Can be a number like 64000000` or a string like ``'64MB'. If None, a single block is used for each file.
                        """,
                    ),
                    'sample': Field(
                        Int,
                        is_required=False,
                        description="Number of bytes to use when determining dtypes.",
                    ),
                    'assume_missing': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If True, all integer columns that aren’t specified in `dtype` are assumed to contain missing values,
                            and are converted to floats. Default is False.
                        """,
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="""
                            Extra options that make sense for a particular storage connection,
                            e.g. host, port, username, password, etc.
                        """,
                    ),
                    'include_path_column': Field(
                        Any,
                        is_required=False,
                        description="""
                            bool or str, Whether or not to include the path to each particular file.
                            If True a new column is added to the dataframe called path.
                            If str, sets new column name. Default is False.
                        """,
                    ),
                }
            ),
            'parquet': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or list, Source directory for data, or path(s) to individual parquet files.
                            Prefix with a protocol like s3:// to read from alternative filesystems.
                            To read from multiple files you can pass a globstring or a list of paths,
                            with the caveat that they must all have the same protocol.
                        """,
                    ),
                    'columns': Field(
                        Any,
                        is_required=False,
                        description="""
                            str or list or None (default), Field name(s) to read in as columns in the output.
                            By default all non-index fields will be read (as determined by the pandas parquet metadata, if present).
                            Provide a single field name instead of a list to read in the data as a Series.
                        """,
                    ),
                    'index': Field(
                        Any,
                        is_required=False,
                        description="""
                            list or False or None (default), Field name(s) to use as the output frame index.
                            By default will be inferred from the pandas parquet file metadata (if present).
                            Use False to read all fields as columns.
                        """,
                    ),
                    'categories': Field(
                        Any,
                        is_required=False,
                        description="""
                            list or dict or None, For any fields listed here,
                            if the parquet encoding is Dictionary, the column will be created with dtype category.
                            Use only if it is guaranteed that the column is encoded as dictionary in all row-groups.
                            If a list, assumes up to 2**16-1 labels; if a dict, specify the number of labels expected;
                            if None, will load categories automatically for data written by dask/fastparquet, not otherwise.
                        """,
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="Key/value pairs to be passed on to the file-system backend, if any.",
                    ),
                    'engine': Field(
                        EngineParquetOptions,
                        is_required=False,
                        description="""
                            Parquet reader library to use.
                            If only one library is installed, it will use that one;
                            if both, it will use ‘fastparquet’.
                        """,
                    ),
                    'gather_statistics': Field(
                        Bool,
                        is_required=False,
                        description="""
                            default is None, Gather the statistics for each dataset partition.
                            By default, this will only be done if the _metadata file is available.
                            Otherwise, statistics will only be gathered if True,
                            because the footer of every file will be parsed (which is very slow on some systems).
                        """,
                    ),
                    'split_row_groups:': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If True (default) then output dataframe partitions will correspond
                            to parquet-file row-groups (when enough row-group metadata is available).
                            Otherwise, partitions correspond to distinct files.
                            Only the “pyarrow” engine currently supports this argument.
                        """,
                    ),
                    'chunksize': Field(
                        Any,
                        is_required=False,
                        description="""
                            int or string, The target task partition size.
                            If set, consecutive row-groups from the same file will be aggregated
                            into the same output partition until the aggregate size reaches this value.
                        """,
                    ),
                }
            ),
            'hdf': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or pathlib.Path or list,
                            File pattern (string), pathlib.Path, buffer to read from, or list of file paths.
                            Can contain wildcards.
                        """,
                    ),
                    'Key': Field(
                        Any,
                        is_required=True,
                        description="group identifier in the store. Can contain wildcards.",
                    ),
                    'start': Field(
                        Int,
                        is_required=False,
                        description="defaults to 0, row number to start at.",
                    ),
                    'stop': Field(
                        Int,
                        is_required=False,
                        description="defaults to None (the last row), row number to stop at.",
                    ),
                    'columns': Field(
                        list,
                        is_required=False,
                        description="A list of columns that if not None, will limit the return columns (default is None).",
                    ),
                    'chunksize': Field(
                        Any,
                        is_required=False,
                        description="Maximal number of rows per partition (default is 1000000).",
                    ),
                    'sorted_index': Field(
                        Bool,
                        is_required=False,
                        description="Option to specify whether or not the input hdf files have a sorted index (default is False).",
                    ),
                    'lock': Field(
                        Bool,
                        is_required=False,
                        description="Option to use a lock to prevent concurrency issues (default is True).",
                    ),
                    'mode': Field(
                        String,
                        is_required=False,
                        description="""
                            {‘a’, ‘r’, ‘r+’}, default ‘a’. Mode to use when opening file(s).
                            ‘r’ - Read-only; no data can be modified.
                            ‘a’ - Append; an existing file is opened for reading and writing, and if the file does not exist it is created.
                            ‘r+’ - It is similar to ‘a’, but the file must already exist.
                        """,
                    ),
                }
            ),
            'json': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or list, Location to read from.
                            If a string, can include a glob character to find a set of file names.
                            Supports protocol specifications such as 's3://'.
                        """,
                    ),
                    'encoding': Field(
                        String,
                        is_required=False,
                        description="The text encoding to implement, e.g., “utf-8”.",
                    ),
                    'errors': Field(
                        String,
                        is_required=False,
                        description="how to respond to errors in the conversion (see str.encode()).",
                    ),
                    'storage_option': Field(
                        Permissive(),
                        is_required=False,
                        description="Passed to backend file-system implementation.",
                    ),
                    'blocksize': Field(
                        Int,
                        is_required=False,
                        description="""
                            default is None, If None, files are not blocked, and you get one partition per input file.
                            If int, which can only be used for line-delimited JSON files,
                            each partition will be approximately this size in bytes, to the nearest newline character.
                        """,
                    ),
                    'sample': Field(
                        Int,
                        is_required=False,
                        description="""
                            Number of bytes to pre-load,
                            to provide an empty dataframe structure to any blocks without data.
                            Only relevant is using blocksize.
                        """,
                    ),
                    'compression': Field(
                        String,
                        is_required=False,
                        description="default is None, String like ‘gzip’ or ‘xz’.",
                    ),
                }
            ),
            'sql_table': Permissive(
                {
                    'table': Field(
                        Any,
                        is_required=True,
                        description="str or sqlalchemy expression, Select columns from here.",
                    ),
                    'uri': Field(
                        String,
                        is_required=True,
                        description="Full sqlalchemy URI for the database connection.",
                    ),
                    'index_col': Field(
                        String,
                        is_required=True,
                        description="""
                        Column which becomes the index, and defines the partitioning.
                        Should be a indexed column in the SQL server, and any orderable type.
                        If the type is number or time, then partition boundaries can be inferred from npartitions or bytes_per_chunk;
                        otherwide must supply explicit divisions=.
                        index_col could be a function to return a value, e.g., sql.func.abs(sql.column('value')).label('abs(value)').
                        index_col=sql.func.abs(sql.column("value")).label("abs(value)"),
                        or index_col=cast(sql.column("id"),types.BigInteger).label("id") to convert the textfield id to BigInteger.
                        Note sql, cast, types methods comes frome sqlalchemy module.
                        Labeling columns created by functions or arithmetic operations is required
                        """,
                    ),
                    'divisions': Field(
                        Any,
                        is_required=False,
                        description="""
                            sequence, Values of the index column to split the table by.
                            If given, this will override npartitions and bytes_per_chunk.
                            The divisions are the value boundaries of the index column used to define the partitions.
                            For example, divisions=list('acegikmoqsuwz') could be used
                            to partition a string column lexographically into 12 partitions,
                            with the implicit assumption that each partition contains similar numbers of records.
                        """,
                    ),
                    'npartitions': Field(
                        Int,
                        is_required=False,
                        description="""
                            Number of partitions, if divisions is not given.
                            Will split the values of the index column linearly between limits, if given, or the column max/min.
                            The index column must be numeric or time for this to work.
                        """,
                    ),
                    'columns': Field(
                        Any,
                        is_required=False,
                        description="""
                            list of strings or None, Which columns to select;
                            if None, gets all; can include sqlalchemy functions,
                            e.g., sql.func.abs(sql.column('value')).label('abs(value)').
                            Labeling columns created by functions or arithmetic operations is recommended.
                        """,
                    ),
                    'bytes_per_chunk': Field(
                        Any,
                        is_required=False,
                        description="""
                            str or int, If both divisions and npartitions is None,
                            this is the target size of each partition, in bytes.
                        """,
                    ),
                    'head_rows': Field(
                        Int,
                        is_required=False,
                        description="How many rows to load for inferring the data-types, unless passing meta.",
                    ),
                    'schema': Field(
                        String,
                        is_required=False,
                        description="""
                            If using a table name, pass this to sqlalchemy to select
                            which DB schema to use within the URI connection.
                        """,
                    ),
                }
            ),
            'table': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or list, Absolute or relative filepath(s).
                            Prefix with a protocol like 's3://' to read from alternative filesystems.
                            To read from multiple files you can pass a globstring or a list of paths,
                            with the caveat that they must all have the same protocol.
                        """,
                    ),
                    'blocksize': Field(
                        Any,
                        is_required=False,
                        description="""
                            str or int or None, Number of bytes by which to cut up larger files.
                            Default value is computed based on available physical memory and the number of cores,
                            up to a maximum of 64MB. Can be a number like 64000000` or a string like ``'64MB'.
                            If None, a single block is used for each file.
                        """,
                    ),
                    'sample': Field(
                        Int,
                        is_required=False,
                        description="Number of bytes to use when determining dtypes.",
                    ),
                    'assume_missing': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If True, all integer columns that aren’t specified in dtype are assumed to contain missing values,
                            and are converted to floats. Default is False.
                        """,
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="""
                            Extra options that make sense for a particular storage connection,
                            e.g. host, port, username, password, etc.
                        """,
                    ),
                    'include_path_column': Field(
                        Any,
                        is_required=False,
                        description="""
                            bool or str, Whether or not to include the path to each particular file.
                            If True a new column is added to the dataframe called path.
                            If str, sets new column name. Default is False.
                        """,
                    ),
                }
            ),
            'fwf': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or list, Absolute or relative filepath(s).
                            Prefix with a protocol like 's3://' to read from alternative filesystems.
                            To read from multiple files you can pass a globstring or a list of paths,
                            with the caveat that they must all have the same protocol.
                        """,
                    ),
                    'blocksize': Field(
                        Any,
                        is_required=False,
                        description="""
                            str or int or None, Number of bytes by which to cut up larger files.
                            Default value is computed based on available physical memory
                            and the number of cores up to a maximum of 64MB.
                            Can be a number like 64000000` or a string like ``'64MB'.
                            If None, a single block is used for each file.
                        """,
                    ),
                    'sample': Field(
                        Int,
                        is_required=False,
                        description="Number of bytes to use when determining dtypes.",
                    ),
                    'assume_missing': Field(
                        Bool,
                        is_required=False,
                        description="""
                            If True, all integer columns that aren’t specified in dtype are assumed
                            to contain missing values, and are converted to floats.
                            Default is False.
                        """,
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="""
                            Extra options that make sense for a particular storage connection,
                            e.g. host, port, username, password, etc.
                        """,
                    ),
                    'include_path_column': Field(
                        Any,
                        is_required=False,
                        description="""
                            bool or str, Whether or not to include the path to each particular file.
                            If True a new column is added to the dataframe called path.
                            If str, sets new column name. Default is False.
                        """,
                    ),
                }
            ),
            'orc': Permissive(
                {
                    'path': Field(
                        Any,
                        is_required=True,
                        description="""
                            str or list, Location of file(s),
                            which can be a full URL with protocol specifier,
                            and may include glob character if a single string.
                        """,
                    ),
                    'columns': Field(
                        list, is_required=False, description="Columns to load. If None, loads all.",
                    ),
                    'storage_options': Field(
                        Permissive(),
                        is_required=False,
                        description="Further parameters to pass to the bytes backend.",
                    ),
                }
            ),
        },
    )
)
def dataframe_loader(_context, config):
    file_type, file_options = list(config.items())[0]
    path = file_options.get('path')

    if file_type == 'csv':
        return dd.read_csv(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'parquet':
        return dd.read_parquet(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'hdf':
        return dd.read_hdf(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'json':
        return dd.read_json(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'sql_table':
        return dd.read_sql_table(**file_options)
    elif file_type == 'table':
        return dd.read_table(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'fwf':
        return dd.read_fwf(path, **dict_without_keys(file_options, 'path'))
    elif file_type == 'orc':
        return dd.read_orc(path, **dict_without_keys(file_options, 'path'))
    else:
        raise DagsterInvariantViolationError(
            'Unsupported file_type {file_type}'.format(file_type=file_type)
        )


def df_type_check(_, value):
    if not isinstance(value, dd.DataFrame):
        return TypeCheck(success=False)
    return TypeCheck(
        success=True,
        metadata_entries=[
            EventMetadataEntry.text(str(len(value)), 'row_count', 'Number of rows in DataFrame'),
            # string cast columns since they may be things like datetime
            EventMetadataEntry.json({'columns': list(map(str, value.columns))}, 'metadata'),
        ],
    )


DataFrame = DagsterType(
    name='DaskDataFrame',
    description='''A Dask DataFrame is a large parallel DataFrame composed of many smaller Pandas DataFrames, split along the index.
    These Pandas DataFrames may live on disk for larger-than-memory computing on a single machine, or on many different machines in a cluster.
    One Dask DataFrame operation triggers many operations on the constituent Pandas DataFrames.
    See https://docs.dask.org/en/latest/dataframe.html''',
    loader=dataframe_loader,
    materializer=dataframe_materializer,
    type_check_fn=df_type_check,
)



import dask
import dask.distributed
from dagster_graphql.client.mutations import execute_execute_plan_mutation

# Dask resource requirements are specified under this key
from dagster import Executor, Field, Permissive, Selector, check, seven
from dagster.core.definitions.executor import check_cross_process_constraints, executor
from dagster.core.events import DagsterEvent
from dagster.core.execution.context.system import SystemPipelineExecutionContext
from dagster.core.execution.plan.plan import ExecutionPlan
from dagster.core.execution.retries import Retries
from dagster.utils import frozentags
from dagster.utils.hosted_user_process import create_in_process_ephemeral_workspace

DASK_RESOURCE_REQUIREMENTS_KEY = 'dagster-dask/resource_requirements'


@executor(
    name='dask',
    config_schema={
        'cluster': Field(
            Selector(
                {
                    'local': Field(
                        Permissive(), is_required=False, description='Local cluster configuration.'
                    ),
                    'yarn': Field(
                        Permissive(), is_required=False, description='YARN cluster configuration.'
                    ),
                    'ssh': Field(
                        Permissive(), is_required=False, description='SSH cluster configuration.'
                    ),
                    'pbs': Field(
                        Permissive(), is_required=False, description='PBS cluster configuration.'
                    ),
                    'moab': Field(
                        Permissive(), is_required=False, description='Moab cluster configuration.'
                    ),
                    'sge': Field(
                        Permissive(), is_required=False, description='SGE cluster configuration.'
                    ),
                    'lsf': Field(
                        Permissive(), is_required=False, description='LSF cluster configuration.'
                    ),
                    'slurm': Field(
                        Permissive(), is_required=False, description='SLURM cluster configuration.'
                    ),
                    'oar': Field(
                        Permissive(), is_required=False, description='OAR cluster configuration.'
                    ),
                    'kube': Field(
                        Permissive(),
                        is_required=False,
                        description='Kubernetes cluster configuration.',
                    ),
                }
            )
        )
    },
)
def dask_executor(init_context):
    '''Dask-based executor.

    The 'cluster' can be one of the following:
    ('local', 'yarn', 'ssh', 'pbs', 'moab', 'sge', 'lsf', 'slurm', 'oar', 'kube').

    If the Dask executor is used without providing executor-specific config, a local Dask cluster
    will be created (as when calling :py:class:`dask.distributed.Client() <dask:distributed.Client>`
    with :py:class:`dask.distributed.LocalCluster() <dask:distributed.LocalCluster>`).

    The Dask executor optionally takes the following config:

    .. code-block:: none

        cluster:
            {
                local?: # takes distributed.LocalCluster parameters
                    {
                        timeout?: 5,  # Timeout duration for initial connection to the scheduler
                        n_workers?: 4  # Number of workers to start
                        threads_per_worker?: 1 # Number of threads per each worker
                    }
            }

    If you'd like to configure a dask executor in addition to the
    :py:class:`~dagster.default_executors`, you should add it to the ``executor_defs`` defined on a
    :py:class:`~dagster.ModeDefinition` as follows:

    .. code-block:: python

        from dagster import ModeDefinition, default_executors, pipeline
        from dagster_dask import dask_executor

        @pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])])
        def dask_enabled_pipeline():
            pass

    '''
    check_cross_process_constraints(init_context)
    ((cluster_type, cluster_configuration),) = init_context.executor_config['cluster'].items()
    return DaskExecutor(cluster_type, cluster_configuration)


def query_on_dask_worker(
    workspace, variables, dependencies, instance_ref=None
):  # pylint: disable=unused-argument
    '''Note that we need to pass "dependencies" to ensure Dask sequences futures during task
    scheduling, even though we do not use this argument within the function.
    '''
    return execute_execute_plan_mutation(workspace, variables, instance_ref=instance_ref)


def get_dask_resource_requirements(tags):
    check.inst_param(tags, 'tags', frozentags)
    req_str = tags.get(DASK_RESOURCE_REQUIREMENTS_KEY)
    if req_str is not None:
        return seven.json.loads(req_str)

    return {}


class DaskExecutor(Executor):
    def __init__(self, cluster_type, cluster_configuration):
        self.cluster_type = check.opt_str_param(cluster_type, 'cluster_type', default='local')
        self.cluster_configuration = check.opt_dict_param(
            cluster_configuration, 'cluster_configuration'
        )

    @property
    def retries(self):
        return Retries.disabled_mode()

    def execute(self, pipeline_context, execution_plan):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
        check.param_invariant(
            isinstance(pipeline_context.executor, DaskExecutor),
            'pipeline_context',
            'Expected executor to be DaskExecutor got {}'.format(pipeline_context.executor),
        )

        # Checks to ensure storage is compatible with Dask configuration
        storage = pipeline_context.run_config.get('storage')
        check.invariant(storage.keys(), 'Must specify storage to use Dask execution')

        check.invariant(
            pipeline_context.instance.is_persistent,
            'Dask execution requires a persistent DagsterInstance',
        )

        # https://github.com/dagster-io/dagster/issues/2440
        check.invariant(
            pipeline_context.system_storage_def.is_persistent,
            'Cannot use in-memory storage with Dask, use filesystem, S3, or GCS',
        )

        step_levels = execution_plan.execution_step_levels()

        pipeline_name = pipeline_context.pipeline_def.name

        instance = pipeline_context.instance

        cluster_type = self.cluster_type
        if cluster_type == 'local':
            from dask.distributed import LocalCluster

            cluster = LocalCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'yarn':
            from dask_yarn import YarnCluster

            cluster = YarnCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'ssh':
            from dask.distributed import SSHCluster

            cluster = SSHCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'pbs':
            from dask_jobqueue import PBSCluster

            cluster = PBSCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'moab':
            from dask_jobqueue import MoabCluster

            cluster = MoabCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'sge':
            from dask_jobqueue import SGECluster

            cluster = SGECluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'lsf':
            from dask_jobqueue import LSFCluster

            cluster = LSFCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'slurm':
            from dask_jobqueue import SLURMCluster

            cluster = SLURMCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'oar':
            from dask_jobqueue import OARCluster

            cluster = OARCluster(**self.build_dict(pipeline_name))
        elif cluster_type == 'kube':
            from dask_kubernetes import KubeCluster

            cluster = KubeCluster(**self.build_dict(pipeline_name))
        else:
            raise ValueError(
                f"Must be providing one of the following ('local', 'yarn', 'ssh', 'pbs', 'moab', 'sge', 'lsf', 'slurm', 'oar', 'kube') not {cluster_type}"
            )

        with dask.distributed.Client(cluster) as client:
            execution_futures = []
            execution_futures_dict = {}

            for step_level in step_levels:
                for step in step_level:
                    # We ensure correctness in sequencing by letting Dask schedule futures and
                    # awaiting dependencies within each step.
                    dependencies = []
                    for step_input in step.step_inputs:
                        for key in step_input.dependency_keys:
                            dependencies.append(execution_futures_dict[key])

                    run_config = dict(pipeline_context.run_config, execution={'in_process': {}})
                    recon_repo = pipeline_context.pipeline.get_reconstructable_repository()
                    variables = {
                        'executionParams': {
                            'selector': {
                                'pipelineName': pipeline_name,
                                'repositoryName': recon_repo.get_definition().name,
                                'repositoryLocationName': '<<in_process>>',
                            },
                            'runConfigData': run_config,
                            'mode': pipeline_context.mode_def.name,
                            'executionMetadata': {'runId': pipeline_context.pipeline_run.run_id},
                            'stepKeys': [step.key],
                        }
                    }

                    dask_task_name = '%s.%s' % (pipeline_name, step.key)

                    workspace = create_in_process_ephemeral_workspace(
                        pointer=pipeline_context.pipeline.get_reconstructable_repository().pointer
                    )

                    future = client.submit(
                        query_on_dask_worker,
                        workspace,
                        variables,
                        dependencies,
                        instance.get_ref(),
                        key=dask_task_name,
                        resources=get_dask_resource_requirements(step.tags),
                    )

                    execution_futures.append(future)
                    execution_futures_dict[step.key] = future

            # This tells Dask to awaits the step executions and retrieve their results to the
            # master
            for future in dask.distributed.as_completed(execution_futures):
                for step_event in future.result():
                    check.inst(step_event, DagsterEvent)

                    yield step_event

    def build_dict(self, pipeline_name):
        '''Returns a dict we can use for kwargs passed to dask client instantiation.

        Intended to be used like:

        with dask.distributed.Client(**cfg.build_dict()) as client:
            << use client here >>

        '''
        if self.cluster_type in ['yarn', 'pbs', 'moab', 'sge', 'lsf', 'slurm', 'oar', 'kube']:
            dask_cfg = {'name': pipeline_name}
        else:
            dask_cfg = {}

        if self.cluster_configuration:
            for k, v in self.cluster_configuration.items():
                dask_cfg[k] = v

        # if address is set, don't add LocalCluster args
        # context: https://github.com/dask/distributed/issues/3313
        if (self.cluster_type == 'local') and ('address' not in dask_cfg):
            # We set threads_per_worker because Dagster is not thread-safe. Even though
            # environments=True by default, there is a clever piece of machinery
            # (dask.distributed.deploy.local.nprocesses_nthreads) that automagically makes execution
            # multithreaded by default when the number of available cores is greater than 4.
            # See: https://github.com/dagster-io/dagster/issues/2181
            # We may want to try to figure out a way to enforce this on remote Dask clusters against
            # which users run Dagster workloads.
            dask_cfg['threads_per_worker'] = 1

        return dask_cfg



import shutil

import dask.dataframe as dd
import pytest
from dagster_dask import DataFrame
from dask.dataframe.utils import assert_eq

from dagster import InputDefinition, OutputDefinition, execute_solid, file_relative_path, solid
from dagster.utils.test import get_temp_dir


def create_dask_df():
    path = file_relative_path(__file__, 'num.csv')
    return dd.read_csv(path)


@pytest.mark.parametrize(
    'file_type,read,kwargs',
    [
        pytest.param('csv', dd.read_csv, {'index': False}, id='csv'),
        pytest.param('parquet', dd.read_parquet, {'write_index': False}, id='parquet'),
        pytest.param('json', dd.read_json, {}, id='json'),
    ],
)
def test_dataframe_outputs(file_type, read, kwargs):
    df = create_dask_df()

    @solid(output_defs=[OutputDefinition(dagster_type=DataFrame, name='output_df')])
    def return_df(_):
        return df

    with get_temp_dir() as temp_path:
        shutil.rmtree(temp_path)
        result = execute_solid(
            return_df,
            run_config={
                'solids': {
                    'return_df': {
                        'outputs': [{'output_df': {file_type: {'path': temp_path, **kwargs}}}]
                    }
                }
            },
        )
        assert result.success
        actual = read(f"{temp_path}/*")
        assert assert_eq(actual, df)


@pytest.mark.parametrize(
    'file_type',
    [
        pytest.param('csv', id='csv'),
        pytest.param('parquet', id='parquet'),
        pytest.param('json', id='json'),
    ],
)
def test_dataframe_inputs(file_type):
    @solid(input_defs=[InputDefinition(dagster_type=DataFrame, name='input_df')])
    def return_df(_, input_df):
        return input_df

    file_name = file_relative_path(__file__, f"num.{file_type}")
    result = execute_solid(
        return_df,
        run_config={
            'solids': {'return_df': {'inputs': {'input_df': {file_type: {'path': file_name}}}}}
        },
    )
    assert result.success
    assert assert_eq(result.output_value(), create_dask_df())



from setuptools import find_packages, setup


def get_version():
    version = {}
    with open('dagster_dask/version.py') as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    return version['__version__']


if __name__ == '__main__':
    setup(
        name='dagster-dask',
        version=get_version(),
        author='Elementl',
        author_email='hello@elementl.com',
        license='Apache-2.0',
        description='Package for using Dask as Dagster\'s execution engine.',
        url='https://github.com/dagster-io/dagster/tree/master/python_modules/libraries/dagster-dask',
        classifiers=[
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        packages=find_packages(exclude=['test']),
        install_requires=[
            'bokeh',
            'dagster',
            'dagster_graphql',
            'dask[dataframe]>=1.2.2',
            'distributed>=1.28.1',
        ],
        extras_require={
            'yarn': ['dask-yarn'],
            'pbs': ['dask-jobqueue'],
            'kube': ['dask-kubernetes'],
        },
        zip_safe=False,
    )



from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION, SUBSCRIPTION_QUERY
from dagster_graphql.implementation.context import DagsterGraphQLContext
from dagster_graphql.schema import create_schema
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector
from graphql import graphql
from graphql.execution.executors.sync import SyncExecutor

from dagster.core.definitions.reconstructable import ReconstructableRepository
from dagster.core.host_representation import InProcessRepositoryLocation
from dagster.core.instance import DagsterInstance
from dagster.utils import file_relative_path


def test_execute_hammer_through_dagit():
    recon_repo = ReconstructableRepository.for_file(
        file_relative_path(__file__, '../../../dagster-test/dagster_test/toys/hammer.py'),
        'hammer_pipeline',
    )
    instance = DagsterInstance.local_temp()

    context = DagsterGraphQLContext(
        locations=[InProcessRepositoryLocation(recon_repo)], instance=instance,
    )

    selector = infer_pipeline_selector(context, 'hammer_pipeline')

    executor = SyncExecutor()

    variables = {
        'executionParams': {
            'runConfigData': {
                'storage': {'filesystem': {}},
                'execution': {'dask': {'config': {'cluster': {'local': {}}}}},
            },
            'selector': selector,
            'mode': 'default',
        }
    }

    start_pipeline_result = graphql(
        request_string=LAUNCH_PIPELINE_EXECUTION_MUTATION,
        schema=create_schema(),
        context=context,
        variables=variables,
        executor=executor,
    )

    if start_pipeline_result.errors:
        raise Exception('{}'.format(start_pipeline_result.errors))

    run_id = start_pipeline_result.data['launchPipelineExecution']['run']['runId']

    context.drain_outstanding_executions()

    subscription = execute_dagster_graphql(context, SUBSCRIPTION_QUERY, variables={'runId': run_id})

    subscribe_results = []
    subscription.subscribe(subscribe_results.append)

    messages = [x['__typename'] for x in subscribe_results[0].data['pipelineRunLogs']['messages']]

    assert 'PipelineStartEvent' in messages
    assert 'PipelineSuccessEvent' in messages
